package org.apache.bcel.generic;

import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.ElementValuePair;
import org.apache.bcel.classfile.AnnotationEntry;
import java.util.Iterator;
import org.apache.bcel.classfile.RuntimeInvisibleParameterAnnotations;
import org.apache.bcel.classfile.RuntimeVisibleParameterAnnotations;
import java.io.IOException;
import org.apache.bcel.classfile.RuntimeInvisibleAnnotations;
import java.io.DataInput;
import org.apache.bcel.classfile.RuntimeVisibleAnnotations;
import java.io.InputStream;
import java.io.DataInputStream;
import java.io.ByteArrayInputStream;
import java.util.ArrayList;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayOutputStream;
import org.apache.bcel.classfile.Attribute;
import java.util.List;

public class AnnotationEntryGen
{
    static final AnnotationEntryGen[] EMPTY_ARRAY;
    private int typeIndex;
    private List<ElementValuePairGen> evs;
    private final ConstantPoolGen cpool;
    private boolean isRuntimeVisible;
    
    static Attribute[] getAnnotationAttributes(final ConstantPoolGen cp, final AnnotationEntryGen[] annotationEntryGens) {
        if (annotationEntryGens.length == 0) {
            return Attribute.EMPTY_ARRAY;
        }
        try {
            int countVisible = 0;
            int countInvisible = 0;
            for (final AnnotationEntryGen a : annotationEntryGens) {
                if (a.isRuntimeVisible()) {
                    ++countVisible;
                }
                else {
                    ++countInvisible;
                }
            }
            final ByteArrayOutputStream rvaBytes = new ByteArrayOutputStream();
            final ByteArrayOutputStream riaBytes = new ByteArrayOutputStream();
            try (DataOutputStream rvaDos = new DataOutputStream(rvaBytes);
                 DataOutputStream riaDos = new DataOutputStream(riaBytes)) {
                rvaDos.writeShort(countVisible);
                riaDos.writeShort(countInvisible);
                for (final AnnotationEntryGen a2 : annotationEntryGens) {
                    if (a2.isRuntimeVisible()) {
                        a2.dump(rvaDos);
                    }
                    else {
                        a2.dump(riaDos);
                    }
                }
            }
            final byte[] rvaData = rvaBytes.toByteArray();
            final byte[] riaData = riaBytes.toByteArray();
            int rvaIndex = -1;
            int riaIndex = -1;
            if (rvaData.length > 2) {
                rvaIndex = cp.addUtf8("RuntimeVisibleAnnotations");
            }
            if (riaData.length > 2) {
                riaIndex = cp.addUtf8("RuntimeInvisibleAnnotations");
            }
            final List<Attribute> newAttributes = new ArrayList<Attribute>();
            if (rvaData.length > 2) {
                newAttributes.add(new RuntimeVisibleAnnotations(rvaIndex, rvaData.length, new DataInputStream(new ByteArrayInputStream(rvaData)), cp.getConstantPool()));
            }
            if (riaData.length > 2) {
                newAttributes.add(new RuntimeInvisibleAnnotations(riaIndex, riaData.length, new DataInputStream(new ByteArrayInputStream(riaData)), cp.getConstantPool()));
            }
            return newAttributes.toArray(Attribute.EMPTY_ARRAY);
        }
        catch (final IOException e) {
            System.err.println("IOException whilst processing annotations");
            e.printStackTrace();
            return null;
        }
    }
    
    static Attribute[] getParameterAnnotationAttributes(final ConstantPoolGen cp, final List<AnnotationEntryGen>[] vec) {
        final int[] visCount = new int[vec.length];
        int totalVisCount = 0;
        final int[] invisCount = new int[vec.length];
        int totalInvisCount = 0;
        try {
            for (int i = 0; i < vec.length; ++i) {
                if (vec[i] != null) {
                    for (final AnnotationEntryGen element : vec[i]) {
                        if (element.isRuntimeVisible()) {
                            final int[] array = visCount;
                            final int n = i;
                            ++array[n];
                            ++totalVisCount;
                        }
                        else {
                            final int[] array2 = invisCount;
                            final int n2 = i;
                            ++array2[n2];
                            ++totalInvisCount;
                        }
                    }
                }
            }
            final ByteArrayOutputStream rvaBytes = new ByteArrayOutputStream();
            try (DataOutputStream rvaDos = new DataOutputStream(rvaBytes)) {
                rvaDos.writeByte(vec.length);
                for (int j = 0; j < vec.length; ++j) {
                    rvaDos.writeShort(visCount[j]);
                    if (visCount[j] > 0) {
                        for (final AnnotationEntryGen element2 : vec[j]) {
                            if (element2.isRuntimeVisible()) {
                                element2.dump(rvaDos);
                            }
                        }
                    }
                }
            }
            final ByteArrayOutputStream riaBytes = new ByteArrayOutputStream();
            try (DataOutputStream riaDos = new DataOutputStream(riaBytes)) {
                riaDos.writeByte(vec.length);
                for (int k = 0; k < vec.length; ++k) {
                    riaDos.writeShort(invisCount[k]);
                    if (invisCount[k] > 0) {
                        for (final AnnotationEntryGen element3 : vec[k]) {
                            if (!element3.isRuntimeVisible()) {
                                element3.dump(riaDos);
                            }
                        }
                    }
                }
            }
            final byte[] rvaData = rvaBytes.toByteArray();
            final byte[] riaData = riaBytes.toByteArray();
            int rvaIndex = -1;
            int riaIndex = -1;
            if (totalVisCount > 0) {
                rvaIndex = cp.addUtf8("RuntimeVisibleParameterAnnotations");
            }
            if (totalInvisCount > 0) {
                riaIndex = cp.addUtf8("RuntimeInvisibleParameterAnnotations");
            }
            final List<Attribute> newAttributes = new ArrayList<Attribute>();
            if (totalVisCount > 0) {
                newAttributes.add(new RuntimeVisibleParameterAnnotations(rvaIndex, rvaData.length, new DataInputStream(new ByteArrayInputStream(rvaData)), cp.getConstantPool()));
            }
            if (totalInvisCount > 0) {
                newAttributes.add(new RuntimeInvisibleParameterAnnotations(riaIndex, riaData.length, new DataInputStream(new ByteArrayInputStream(riaData)), cp.getConstantPool()));
            }
            return newAttributes.toArray(Attribute.EMPTY_ARRAY);
        }
        catch (final IOException e) {
            System.err.println("IOException whilst processing parameter annotations");
            e.printStackTrace();
            return null;
        }
    }
    
    public static AnnotationEntryGen read(final DataInput dis, final ConstantPoolGen cpool, final boolean b) throws IOException {
        final AnnotationEntryGen a = new AnnotationEntryGen(cpool);
        a.typeIndex = dis.readUnsignedShort();
        for (int elemValuePairCount = dis.readUnsignedShort(), i = 0; i < elemValuePairCount; ++i) {
            final int nidx = dis.readUnsignedShort();
            a.addElementNameValuePair(new ElementValuePairGen(nidx, ElementValueGen.readElementValue(dis, cpool), cpool));
        }
        a.isRuntimeVisible(b);
        return a;
    }
    
    public AnnotationEntryGen(final AnnotationEntry a, final ConstantPoolGen cpool, final boolean copyPoolEntries) {
        this.cpool = cpool;
        if (copyPoolEntries) {
            this.typeIndex = cpool.addUtf8(a.getAnnotationType());
        }
        else {
            this.typeIndex = a.getAnnotationTypeIndex();
        }
        this.isRuntimeVisible = a.isRuntimeVisible();
        this.evs = this.copyValues(a.getElementValuePairs(), cpool, copyPoolEntries);
    }
    
    private AnnotationEntryGen(final ConstantPoolGen cpool) {
        this.cpool = cpool;
    }
    
    public AnnotationEntryGen(final ObjectType type, final List<ElementValuePairGen> elements, final boolean vis, final ConstantPoolGen cpool) {
        this.cpool = cpool;
        this.typeIndex = cpool.addUtf8(type.getSignature());
        this.evs = elements;
        this.isRuntimeVisible = vis;
    }
    
    public void addElementNameValuePair(final ElementValuePairGen evp) {
        if (this.evs == null) {
            this.evs = new ArrayList<ElementValuePairGen>();
        }
        this.evs.add(evp);
    }
    
    private List<ElementValuePairGen> copyValues(final ElementValuePair[] in, final ConstantPoolGen cpool, final boolean copyPoolEntries) {
        final List<ElementValuePairGen> out = new ArrayList<ElementValuePairGen>();
        for (final ElementValuePair nvp : in) {
            out.add(new ElementValuePairGen(nvp, cpool, copyPoolEntries));
        }
        return out;
    }
    
    public void dump(final DataOutputStream dos) throws IOException {
        dos.writeShort(this.typeIndex);
        dos.writeShort(this.evs.size());
        for (final ElementValuePairGen envp : this.evs) {
            envp.dump(dos);
        }
    }
    
    public AnnotationEntry getAnnotation() {
        final AnnotationEntry a = new AnnotationEntry(this.typeIndex, this.cpool.getConstantPool(), this.isRuntimeVisible);
        for (final ElementValuePairGen element : this.evs) {
            a.addElementNameValuePair(element.getElementNameValuePair());
        }
        return a;
    }
    
    public int getTypeIndex() {
        return this.typeIndex;
    }
    
    public final String getTypeName() {
        return this.getTypeSignature();
    }
    
    public final String getTypeSignature() {
        final ConstantUtf8 utf8 = (ConstantUtf8)this.cpool.getConstant(this.typeIndex);
        return utf8.getBytes();
    }
    
    public List<ElementValuePairGen> getValues() {
        return this.evs;
    }
    
    public boolean isRuntimeVisible() {
        return this.isRuntimeVisible;
    }
    
    private void isRuntimeVisible(final boolean b) {
        this.isRuntimeVisible = b;
    }
    
    public String toShortString() {
        final StringBuilder s = new StringBuilder();
        s.append("@").append(this.getTypeName()).append("(");
        for (int i = 0; i < this.evs.size(); ++i) {
            s.append(this.evs.get(i));
            if (i + 1 < this.evs.size()) {
                s.append(",");
            }
        }
        s.append(")");
        return s.toString();
    }
    
    @Override
    public String toString() {
        final StringBuilder s = new StringBuilder(32);
        s.append("AnnotationGen:[").append(this.getTypeName()).append(" #").append(this.evs.size()).append(" {");
        for (int i = 0; i < this.evs.size(); ++i) {
            s.append(this.evs.get(i));
            if (i + 1 < this.evs.size()) {
                s.append(",");
            }
        }
        s.append("}]");
        return s.toString();
    }
    
    static {
        EMPTY_ARRAY = new AnnotationEntryGen[0];
    }
}
